#
# lutron.device - Lutron protocol devices
#
# A Lutron installation links together a set of Lutron devices.
# One of them is the *Controller Interface* device which we talk to.
# Others are hardware pieces such as shade motors and switches.
# Here we characterize them.
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import asyn.utility

import time

import core


#
# Lutron Action numbers used here
#
DEVICE_PRESS =	3		# button press (even virtual ones)
DEVICE_MAIN =	14		# master setting (brightness, shade position, ...)

# interactive shade motion commands
SHADE_RAISE =	18		# raise, open, expose
SHADE_LOWER =	19		# lower, close, cover
SHADE_STOP =	20		# stop wherever you are
SHADE_MOTION =	21		# [undocumented] motion state (moving if nonzero)

VENETIAN_LIFT_TILT = 23	# venetian blinds left and tilt compound command
VENETIAN_LIFT =	24		# venetian blinds lift
VENETIAN_TILT =	25		# venetian blinds tilt

# programming commands
PROG_ADD_SHADE = 1		# add a shade member
PROG_CLEAR =	3		# clear all programming from phantom button
PROG_ADD_VENETIAN = 5 	# add venetian blinds (both lift and tilt)
PROG_ADD_TILT = 6		# add venetian blinds (tilt only)
PROG_ADD_LIFT = 7		# add venetian blinds (lift only)


#
# Device types and creation
#
DEVICETYPES = { }		# name -> Device subclass
DEVICEMATCH = []		# class search order

def makedev(interface, desc):
	""" Given a "Details" description, create the right kind of device object. """
	for cls in DEVICEMATCH:
		if cls._match(desc):
			return cls(interface, desc)


#
# Base class of Lutron devices.
# You must use a subclass for actual devices.
#
class Device(asyn.Callable):
	""" A Lutron Device, as discovered by an Interface. """

	class __metaclass__(type):
		def __init__(cls, name, bases, content):
			type.__init__(cls, name, bases, content)
			# record in class map unless private
			if name[0] != '_':  # not private
				DEVICETYPES[name] = cls
				if 'MATCH' in content:
					DEVICEMATCH.insert(0, cls)

	NAME = 'Device'		# display name of this device class

	def __init__(self, interface, desc):
		asyn.Callable.__init__(self)
		self.interface = interface
		self.family = desc["FAMILY"]	# product family
		self.product = desc["PRODUCT"]	# product code
		self.serial = int(desc["SN"], 16) # serial number (hex)
		iid = desc["INTEGRATIONID"]		# integration id (i.e. name)
		self.iid = None if iid == '(Not Set)' else iid # (make it None if not set)
		self.states = { }				# Lutron-reported states (by action code)
		self.inButtons = set()			# Phantom Buttons we're part of
		self._discover()

	def _discover(self):
		self.command("?DEVICE", 0, 0)	# tell me about your current state

	@property
	def value(self):
		return self.states.get(DEVICE_MAIN)

	@property
	def motion(self):
		return self.states.get(SHADE_MOTION)

	@property
	def description(self):
		""" This device's Integration ID, or its serial number if we don't have one. """
		return self.iid or "0x%x" % self.serial

	@classmethod
	def _match(cls, desc):
		""" See if a Details description describes this kind of device. """
		for key, value in cls.MATCH.items():
			if key not in desc or desc[key] != value:
				return False
		return True

	def refresh(self):
		self.command("?DEVICE", 0, 0)

	def update(self, comp, action, value):
		""" Update state from incoming report. """
		assert comp == 0	# no components at this level
		if action == DEVICE_MAIN:
			# values are physical data; round to nearest int
			value = int(round(float(value)))
			self.states[DEVICE_MAIN] = value
			if not self.interface._discovering:
				self.callout('value', value)
		elif action == SHADE_MOTION:
			value = float(value)
			self.states[SHADE_MOTION] = value
			self.callout('motion', value)
		else:
			self.states[action] = value
		for button in self.inButtons:
			button._memberUpdate(self)

	def _discovery_done(self):
		""" Sent when all devices are processed. Hook to manage circular dependencies. """
		pass

	def command(self, cmd, *args):
		""" Relay (cmd, args) to interface as (cmd, my-id, args) """
		self.interface.command(cmd, hex(self.serial), *args)

	def __repr__(self):
		return "<%s:%s/%s sn=0x%08.8x id=%s value=%s>" % (
			type(self).__name__, self.family, self.product, self.serial, self.iid, self.value
		)


#
# Catch-all for unrecognized kinds of devices.
# We don't support them, so we do our best to ignore them.
#
class Unknown(Device):
	MATCH = { }
	NAME = 'Unknown'


#
# Shade devices.
# This is the common features of shades, blinds, and such.
#
class Shades(Device):
	""" A Devices of the Shades family. """

	@property
	def value(self):
		return self.states.get(DEVICE_MAIN)

	@property
	def moving(self):
		return self.states.get(SHADE_MOTION)


#
# Simple roller shades. Up or down; left or right; that's it.
#
class RollerShades(Shades):
	""" A "plain" Shades device. It rolls up and down, and that's it.

		This maps perfectly onto a standard dimmer device.
		It also supports moveTo (but ignores any tilt).
	"""
	MATCH = dict(FAMILY='SHADES(3)', PRODUCT='ROLLER(1)')
	NAME = 'Roller Shades'

	def move(self, type):
		""" Issue an interactive motion command. """
		self.command("#DEVICE", 0, type)

	def moveTo(self, lift, tilt=None):
		""" Roller shades moveTo. Ignores tilt entirely. """
		self.command("#DEVICE", 0, DEVICE_MAIN, float(lift))


#
# Venetian blinds. Move slats up and down AND tilt them.
#
class VenetianBlinds(Shades):
	""" A Venetian Blinds device. It moves up and down and tilts. """

	MATCH = dict(FAMILY='SHADES(3)', PRODUCT='VENETIAN(3)')
	NAME = 'Venetian Blinds'

	def moveTo(self, lift, tilt):
		if lift is not None:
			if tilt is not None:
				self.command("#DEVICE", 0, VENETIAN_LIFT_TILT, int(lift), int(tilt))
			else:
				self.command("#DEVICE", 0, VENETIAN_LIFT, int(lift))
		else:
			if tilt is not None:
				self.command("#DEVICE", 0, VENETIAN_TILT, int(tilt))


#
# A control interface device.
#
class ControlInterface(Device):
	""" A control interface. It talks to us, and has phantom buttons that can be programmed. """
	MATCH = dict(FAMILY='CONTROL_INTERFACE(6)', PRODUCT='QSE(1)')
	NAME = 'Control Interface'

	def __init__(self, interface, desc):
		super(ControlInterface, self).__init__(interface, desc)
		interface.nwk = self	# record NWK - presume the only one
		self.phantoms = { }

	def _discover(self):
		self.command("?PROGRAMMING")	# tell me about your buttons

	def phantom(self, button):
		""" Create and return a phantom button by number. """
		button = int(button)
		if button not in self.phantoms:
			self.phantoms[button] = PhantomButton(self, button)
		return self.phantoms[button]

	def update(self, button, action, *args):
		""" Process device feedback for phantom buttons. """
		# buttons just get blind echoes of commands issued to them. We'll synthesize state ourselves

	def updateProgram(self, button, dev, subdev, lift, tilt):
		""" Update button programming based on incoming data. """
		assert int(subdev) == 0		# shades and blinds have no subdevices
		self.phantom(button)._programAdd(dev, subdev, lift, tilt)

	def clearProgram(self, button):
		self.phantom(button)._programClear()
		del self.phantoms[button]

	def _discovery_done(self):
		super(ControlInterface, self)._discovery_done()
		for button in self.phantoms.values():
			button._memberUpdate(None)				# set initial values


#
# A "phantom button" on a control interface device.
#
class PhantomButton(asyn.Callable):
	""" A Phantom Button on a ControlInterface device.

		This conveniently duck-types like a Device, though it isn't.
		Note that Lutron doesn't provide summary device-like state for these buttons;
		we're making those up.
	"""
	def __init__(self, nwk, bn, callout=None):
		asyn.Callable.__init__(self, callout=callout)
		self.nwk = nwk
		self.bn = bn
		self.interface = nwk.interface
		self.links = { }
		self.min_mode = False					# value state is max of devices
		self.value = None
		self.motion = 0

	@property
	def description(self):
		return "%s button %d" % (self.nwk.description, self.bn)
	
	def hasDeviceType(self, cls):
		for dev in self.links:
			if isinstance(dev, cls):
				return True


	#
	# Commanded movements
	#
	def press(self):
		""" Virtually press this phantom button to move to its presets. """
		self.command("#DEVICE", DEVICE_PRESS)

	def moveTo(self, lift=None, tilt=None):
		""" Tell devices programmed to this button to move to explicit shared positions. """
		device = self.nwk
		if lift is not None:
			if tilt is not None:
				# order matters: roller commands override blinds but not vice versa
				if self.hasDeviceType(RollerShades):
					self.command("#DEVICE", DEVICE_MAIN, int(lift))
				if self.hasDeviceType(VenetianBlinds):
					self.command("#DEVICE", VENETIAN_LIFT_TILT, int(lift), int(tilt))
			else:
				self.command("#DEVICE", DEVICE_MAIN, int(lift))	# (VENETIAN_LIFT applies to blinds only)
		else:
			if tilt is not None:
				self.command("#DEVICE", VENETIAN_TILT, int(tilt))

	#
	# Commanded updates
	#
	def programAdd(self, dev, lift=None, tilt=None):
		if not isinstance(dev, Device):
			dev = self.interface.device(dev)
		def prog(how, *args):
			self.command("#PROGRAMMING", how,
				dev.description, 0, *map(lambda s: '' if s is None else s, args))
		if lift is not None:
			if tilt is not None:
				prog(PROG_ADD_VENETIAN, lift, tilt)
			else:
				prog(PROG_ADD_LIFT, lift)
		else:
			if tilt is not None:
				prog(PROG_ADD_TILT, tilt)
		self.command("?PROGRAMMING")

	def programClear(self):
		self.command("#PROGRAMMING", PROG_CLEAR)

	#
	# Feedback updates
	#
	def _memberUpdate(self, member):
		""" Update our state due to a change to a member device.
		
			Self.value will be the max (highest, most open) of its members,
			unless self.min_mode is True, in which case it'll be the most closed one.
			Self.motion is the longest of the member motion values, i.e. the time for
			the last one to arrive at its destination.

			Member may be None for a general update.
		"""
		value = 100 if self.min_mode else 0
		motion = 0
		old_motion = self.motion
		for dev in self.links:
			value = min(value, dev.value) if self.min_mode else max(value, dev.value)
			motion = max(motion, dev.motion)
		self.value = value
		self.motion = motion
		self.callout('value', value)
		if motion != old_motion:
			self.callout('motion', motion)

	def _programAdd(self, dev, subdev, lift, tilt):
		assert subdev == 0
		dev = self.interface.device(dev)
		self.links[dev] = (lift, tilt)
		dev.inButtons.add(self)

	def _programClear(self):
		for dev in self.links:
			dev.inButtons.discard(self)
		self.links = { }
	
	#
	# Button commands (go to interface, add button number)
	#
	def command(self, cmd, *args):
		""" Send a command to a particular button of our nwk. """
		self.nwk.command(cmd, self.bn, *args)


#
# Keypads (stub)
#
class Keypad(Device):
	MATCH = dict(FAMILY='KEYPAD(1)')
	NAME = 'Keypad'

	def update(self, comp, action, *args):
		print self, 'UPDATE', comp, action, args
